#ifndef TTIME_H
#define TTIME_H
#include <iostream>
#include <iomanip>

#include <sys/types.h>
#include <time.h>
#include <sys/time.h>


class Time {
    private:
	struct timeval tv;
    public:
	Time(time_t t=0): tv() { tv.tv_sec = t; tv.tv_usec = 0; }
	Time(int t): tv() { tv.tv_sec = t; tv.tv_usec = 0; }
	Time(const struct timeval& t) : tv(t) { }
	Time(double d) : tv() { 
	    tv.tv_sec = static_cast<int>(d); 
	    tv.tv_usec = static_cast<int>((d - tv.tv_sec) * 1000000); 
	}
	Time(const Time &t) : tv(t.tv) { }
	double d_val() const { return tv.tv_sec + tv.tv_usec / 1e6; }
	struct timeval tv_val() const { return tv; }
	time_t t_val() const { return static_cast<time_t>(tv.tv_sec); }
	struct tm tm_val() const {
	    struct tm *t = localtime(&tv.tv_sec); 
	    return *t;
	}
	struct tm gtm_val() const {
	    struct tm *t = gmtime(&tv.tv_sec); 
	    return *t;
	}
	virtual ostream& print(ostream& f) const {
	    char fi = f.fill();
	    f.fill('0');
	    f << tv.tv_sec << "." << setw(6) << tv.tv_usec;
	    f.fill(fi);
	    return f;

	}
	bool operator<(const Time& t) const { 
	    if ( tv.tv_sec != t.tv.tv_sec ) return tv.tv_sec < t.tv.tv_sec;
	    else return tv.tv_usec < t.tv.tv_usec;
	}
	bool operator>(const Time& t) const { 
	    if ( tv.tv_sec != t.tv.tv_sec ) return tv.tv_sec > t.tv.tv_sec;
	    else return tv.tv_usec > t.tv.tv_usec;
	}
	bool operator==(const Time& t) const { 
	    return tv.tv_sec == t.tv.tv_sec && tv.tv_usec == t.tv.tv_usec;
	}
	bool operator<=(const Time& t) const {
	    return operator<(t) || operator==(t);
	}
	bool operator>=(const Time& t) const {
	    return operator>(t) || operator==(t);
	}
	// Assignment operators to allow times to be assigned to as well as
	// initialized.
	Time& operator=(time_t t) { 
	    tv.tv_sec = t; tv.tv_usec = 0;
	    return (*this); 
	}
	Time& operator=(int t) { 
	    tv.tv_sec = t; tv.tv_usec = 0;
	    return (*this); 
	}
	Time& operator=(const struct timeval& t) { tv = t; return *this; }
	Time& operator=(double d) {
	    tv.tv_sec = static_cast<int>(d); 
	    tv.tv_usec = static_cast<int>((d - tv.tv_sec) * 1000000); 
	    return *this;
	}
	Time& operator=(const Time &t) { tv = t.tv; return *this; }

	// Helpers will make these ops more useful
	Time operator+(const Time& t) const {
	    struct timeval nt;

	    nt.tv_sec = (tv.tv_sec + t.tv.tv_sec ) + 
		(tv.tv_usec + t.tv.tv_usec) / 1000000;
	    nt.tv_usec = (tv.tv_usec + t.tv.tv_usec) % 1000000;
	    return Time(nt);
	}

	Time operator-(const Time& t) const {
	    struct timeval nt;

	    nt.tv_sec = tv.tv_sec - t.tv.tv_sec ;
	    if ( tv.tv_usec >= t.tv.tv_usec ) 
		nt.tv_usec = tv.tv_usec - t.tv.tv_usec;
	    else {
		nt.tv_sec--;
		nt.tv_usec = 1000000 + tv.tv_usec - t.tv.tv_usec;
	    }
	    return Time(nt);
	}

	Time operator*(int x) const {
	    long us = x * tv.tv_usec;
	    struct timeval nt;

	    nt.tv_usec = us % 1000000;
	    nt.tv_sec = x * tv.tv_sec + (us / 1000000);
	    return Time(nt);
	}
	Time operator*(double x) const {
	    double us = x * tv.tv_usec;
	    struct timeval nt;

	    nt.tv_usec = static_cast<long>(us - (us/1000000)*1000000);
	    nt.tv_sec = static_cast<long>(x * tv.tv_sec)
		+ static_cast<long>(us / 1000000);
	    return Time(nt);
	}
	// Division is done as a double, because integer division doesn't
	// distribute.
	Time operator/(int x) const { return Time(d_val()/x); }
	Time operator/(double x) const { return Time(d_val()/x); }

	Time& operator+=(const Time& t) { return operator=(operator+(t)); }
	Time& operator-=(const Time& t) { return operator=(operator-(t)); }
	Time& operator*=(int x) { return operator=(operator*(x)); }
	Time& operator*=(double x) { return operator=(operator*(x)); }
	Time& operator/=(int x) { return operator=(operator/(x)); }
	Time& operator/=(double x) { return operator=(operator/(x)); }
};

inline ostream& operator<<(ostream& f, const Time &t) { 
    return t.print(f); 
}

// These allow promotion to do the right thing
inline Time operator+(time_t i, const Time& t) { return t + i; }
inline Time operator+(double d, const Time& t) { return t + d; }
inline Time operator+(const struct timeval tv, const Time& t) { return t + tv; }
inline Time operator-(time_t i, const Time& t) { return Time(i) - t; }
inline Time operator-(double d, const Time& t) { return Time(d) - t; }
inline Time operator-(const struct timeval tv, const Time& t) { 
    return Time(tv) - t;
}
// These allows commutation of the multiples
inline Time operator*(int x, const Time& t) { return t * x; }
inline Time operator*(double x, const Time& t) { return t * x; }

// Get the current time as a Time
inline Time now() { 
    struct timeval tv;
    gettimeofday(&tv, 0);
    return Time(tv); 
}

#endif
